import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.compute.compiled import collect_cython
import numpy

def test_collector():

    columns = collect_cython([(1, 2), (2, 1), (7, 8)], numpy.array([1, 0], dtype=numpy.int32))
    assert len(columns) == 2
    assert len(columns[0]) == 3
    assert sum(columns[0]) == 11
    assert sum(columns[1]) == 10

if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests
    test_collector()
    run_tests()



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from tests import cities


def test_can_profile():
    # fmt:off
    
    import orso
    df = orso.DataFrame(cities.values)
    profile = df.profile.to_dataframe()

    assert profile.shape == (6, 12), profile.shape
    assert profile.collect("count") == [20] * 6


def test_opteryx_profile_planets():
    try:
        sys.path.insert(1, os.path.join(sys.path[0], "../../opteryx"))
        import opteryx

        planets = opteryx.query("SELECT * FROM $planets")
        profile = planets.profile.to_dataframe()
        assert profile.shape == (20, 12), profile.shape
        assert profile.collect("count") == [9] * 20
    except ImportError:
        # if Opteryx isn't installed, don't fail
        pass


def test_opteryx_profile_satellites():
    try:
        sys.path.insert(1, os.path.join(sys.path[0], "../../opteryx"))
        import opteryx

        planets = opteryx.query("SELECT * FROM $satellites")
        profile = planets.profile.to_dataframe()
        assert profile.shape == (8, 12), profile.shape
        assert profile.collect("count") == [177] * 8
    except ImportError:
        # if Opteryx isn't installed, don't fail
        pass


def test_opteryx_profile_astronauts():
    try:
        sys.path.insert(1, os.path.join(sys.path[0], "../../opteryx"))
        import opteryx

        planets = opteryx.query("SELECT * FROM $astronauts")
        profile = planets.profile.to_dataframe()
        assert profile.shape == (19, 12), profile.shape
        assert profile.collect("count") == [357] * 19
    except ImportError:
        # if Opteryx isn't installed, don't fail
        pass


def test_opteryx_profile_missions():
    try:
        sys.path.insert(1, os.path.join(sys.path[0], "../../opteryx"))
        import opteryx

        missions = opteryx.query("SELECT * FROM $missions")
        profile = missions.profile.to_dataframe()
        assert profile.shape == (8, 12), profile.shape
        assert profile.collect("count") == [4630] * 8, profile.collect("count")
    except ImportError:
        # if Opteryx isn't installed, don't fail
        pass


def test_opteryx_profile_fake():
    try:
        sys.path.insert(1, os.path.join(sys.path[0], "../../opteryx"))
        import opteryx

        planets = opteryx.query("SELECT * FROM FAKE(100, 100) AS FK")
        profile = planets.profile.to_dataframe()
        assert profile.shape == (100, 12), profile.shape
        assert profile.collect("count") == [100] * 100
    except ImportError:
        # if Opteryx isn't installed, don't fail
        pass


def test_profile_estimators():
    """
    Make sure the estimators are returning something like what
    they're meant to.

    Build a profile, and then test against results from SQL queries
    which perform the same analysis
    """
    sys.path.insert(1, os.path.join(sys.path[0], "../../opteryx"))
    import opteryx
    from orso.profiler import TableProfile

    missions = opteryx.query("SELECT * FROM $missions")
    profile: TableProfile = missions.profile

    source = opteryx.query("SELECT COUNT(*) as missing FROM $missions WHERE Lauched_at IS NULL")
    values = source.fetchone().as_dict
    assert (
        profile.column("Lauched_at").missing == values["missing"]
    ), f"{profile.column('Lauched_at').missing}, {values['missing']}"

    source = opteryx.query("SELECT MIN(Price) as minimum FROM $missions")
    values = source.fetchone().as_dict
    assert profile.column("Price").minimum == int(
        values["minimum"]
    ), f"{profile.column('Price').minimum} != {values['minimum']}"

    source = opteryx.query("SELECT MAX(Price) as maximim FROM $missions")
    values = source.fetchone().as_dict
    assert profile.column("Price").maximum == int(
        values["maximim"]
    ), f"{profile.column('Price').maximum} != {values['maximim']}"

    source = opteryx.query(
        "SELECT COUNT(*) AS frequency, Company FROM $missions GROUP BY Company ORDER BY COUNT(*) DESC"
    )
    values = source.fetchone().as_dict
    assert profile.column("Company").most_frequent_values[0] == values["Company"], values
    assert profile.column("Company").most_frequent_counts[0] == values["frequency"], values

    source = opteryx.query("SELECT COUNT_DISTINCT(Lauched_at) AS unique_timestamps FROM $missions")
    values = source.fetchone().as_dict
    estimated_cardinality = profile.column("Lauched_at").estimate_cardinality()
    assert (
        estimated_cardinality * 0.75 < values["unique_timestamps"] < estimated_cardinality * 1.25
    ), f"{profile.column('Launched_at').estimate_cardinality()} != {values['unique_timestamps']}"

    source = opteryx.query("SELECT COUNT(*) AS price_over_100 FROM $missions WHERE Price > 100")
    values = source.fetchone().as_dict
    estimate = profile.column("Price").estimate_values_above(100)
    assert (
        estimate * 0.9 < values["price_over_100"] < estimate * 1.1
    ), f"{estimate}, {values['price_over_100']}"

    source = opteryx.query("SELECT COUNT(*) AS price_under_250 FROM $missions WHERE Price < 250")
    values = source.fetchone().as_dict
    estimate = profile.column("Price").estimate_values_below(250)
    assert (
        estimate * 0.9 < values["price_under_250"] < estimate * 1.1
    ), f"{estimate}, {values['price_under_250']}"


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ---------------------------------------------------------------------------------
# Our record when converted to bytes looks like this
#
#       ┌───────────┬─────────────┬──────────────┐
#       │   flags   │   row_size  │  row values  │
#       └───────────┴─────────────┴──────────────┘
#
# 'flags' currently has a a version string and a single flag of deleted.
#
# Row object looks and acts like a Tuple where possible, but has additional features
# such as as_dict() to render as a dictionary.

import datetime
import time
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import numpy
import orjson
from ormsgpack import OPT_SERIALIZE_NUMPY
from ormsgpack import packb

from orso.compute.compiled import extract_dict_columns
from orso.compute.compiled import from_bytes_cython
from orso.exceptions import DataError
from orso.schema import RelationSchema

HEADER_SIZE: int = 14
HEADER_PREFIX: bytes = b"\x10\x00"
MAXIMUM_RECORD_SIZE: int = 16 * 1024 * 1024


def extract_columns(table: List[Dict[str, Any]], columns: List[str]) -> Tuple[List[Any], ...]:
    """
    Extracts specified columns from a table.

    Parameters:
        table: List of dictionaries representing rows in a table.
        columns: List of column names to extract.
    Returns:
        A tuple of lists, each containing values of a particular column.
    """
    n_rows = len(table)
    result = [None] * len(columns)
    for i in range(len(columns)):
        result[i] = [None] * n_rows  # type:ignore

    for i, row in enumerate(table):
        for j, column in enumerate(columns):
            result[j][i] = row[column]  # type:ignore

    return tuple(result)


class Row(tuple):
    __slots__ = ()
    _fields: Tuple[str, ...] = None
    _cached_map: Tuple[Tuple[str, Any]] = None
    _cached_byte_size: int = None
    _key: Tuple[int, int] = None

    def __new__(cls, data: Union[Dict[str, Any], Tuple[Any, ...]]):
        """
        Creates a new Row instance.

        Parameters:
            data: Either a dictionary or a tuple.
        Returns:
            A new Row instance.
        """
        if isinstance(data, dict):
            # data = tuple([data.get(field) for field in cls._fields])
            # previous comments on the below line suggested it had a bug, but didn't
            # say what the bug was - this is about 25% faster than the pure Python version
            # There is a lot of testing on this function and it hasn't found any bugs.
            data = extract_dict_columns(data, cls._fields)  # type:ignore
        instance = super().__new__(cls, data)  # type:ignore
        return instance

    def get(self, item, default=None):
        index = self._fields.index(item)
        if index == -1:
            return default
        return self[index]

    @property
    def as_map(self) -> Tuple[Tuple[str, Any], ...]:
        """
        Returns the Row as a tuple of key-value pair tuples (a 'map').

        Returns:
            A tuple of key-value pair tuples.
        """
        if self._cached_map is None:
            self._cached_map = tuple(zip(self._fields, self))  # type:ignore
        return self._cached_map

    @property
    def as_dict(self) -> Dict[str, Any]:
        """
        Returns the Row as a dictionary.

        Returns:
            A dictionary representation of the Row.
        """
        return dict(self.as_map)

    @property
    def values(self) -> Tuple[Any, ...]:
        return tuple(self)

    def keys(self) -> Tuple[str, ...]:
        return self._fields

    @classmethod
    def from_bytes(cls, data: bytes) -> "Row":
        """
        Creates a Row instance from bytes.

        Parameters:
            data: The byte representation of a Row.
        Returns:
            A new Row instance.
        """
        return cls(from_bytes_cython(data))

    def nbytes(self) -> int:
        if self._cached_byte_size is None:
            self._cached_byte_size = len(self.as_bytes)
        return self._cached_byte_size

    @property
    def as_bytes(self) -> bytes:
        """
        Converts the Row to bytes.

        Returns:
            The byte representation of the Row.
        """

        def serialize(value):
            if isinstance(value, numpy.datetime64):
                if numpy.isnat(value):
                    return None
                return ("__datetime__", value.astype("datetime64[s]").astype("int"))
            if isinstance(value, datetime.datetime):
                return ("__datetime__", value.timestamp())
            if isinstance(value, numpy.ndarray):
                return value.tolist()
            return str(value)

        record_bytes = packb(tuple(self), option=OPT_SERIALIZE_NUMPY, default=serialize)
        record_size = len(record_bytes)
        timestamp = time.time_ns()

        if record_size > MAXIMUM_RECORD_SIZE:
            raise DataError("Record length cannot exceed 16Mb")

        return (
            HEADER_PREFIX
            + record_size.to_bytes(4, "big")
            + timestamp.to_bytes(8, "big")
            + record_bytes
        )

    @property
    def as_json(self) -> bytes:
        """
        Converts the Row to JSON bytes.

        Returns:
            The JSON byte representation of the Row.
        """
        return orjson.dumps(self.as_dict, default=str)

    @classmethod
    def create_class(
        cls, schema: Union[RelationSchema, Tuple[str, ...], List[str]], tuples_only: bool = False
    ) -> type:
        """
        Creates a subclass of Row based on the schema provided.

        Parameters:
            schema: Either a RelationSchema or a list/tuple of field names.
        Returns:
            A new Row subclass.
        """
        if isinstance(schema, RelationSchema):
            fields = tuple(c.name for c in schema.columns)
        elif not isinstance(schema, (list, tuple)):
            raise ValueError("Row requires either a list of field names or a RelationSchema")

        fields = tuple(str(s) for s in schema)
        if tuples_only:
            # if we're only handling tuples, we can delegate to super
            return type("RowFactory", (Row,), {"_fields": fields, "__new__": super().__new__})
        return type("RowFactory", (Row,), {"_fields": fields})



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

import pytest

from orso.bitarray import BitArray


def test_init_size():
    with pytest.raises(AssertionError, match="bitarray size must be a positive integer"):
        BitArray(0)


def test_get_index_error():
    b = BitArray(8)
    with pytest.raises(IndexError, match="Index out of range"):
        print(b.get(8))
    with pytest.raises(IndexError, match="Index out of range"):
        print(b.get(-1))


def test_set_index_error():
    b = BitArray(8)
    with pytest.raises(IndexError, match="Index out of range"):
        b.set(8, 1)
    with pytest.raises(IndexError, match="Index out of range"):
        b.set(-1, 1)


def test_get_set():
    b = BitArray(8)
    b.set(2, 1)
    assert b.get(2) is True
    b.set(2, 0)
    assert b.get(2) is False


def test_bits_representation():
    b = BitArray(8)
    b.set(2, 1)
    b.set(4, 1)
    assert b.array == bytearray([20]), b.array


def test_load_and_unload():
    b = BitArray(12)
    b.set(1, 1)
    b.set(5, 1)
    b.set(11, 1)

    d = BitArray.from_array(b.array, 12)

    assert b.array.hex() == d.array.hex()

    assert d.get(0) == 0
    assert d.get(1) == 1
    assert d.get(2) == 0
    assert d.get(3) == 0
    assert d.get(4) == 0
    assert d.get(5) == 1
    assert d.get(6) == 0
    assert d.get(7) == 0
    assert d.get(8) == 0
    assert d.get(9) == 0
    assert d.get(10) == 0
    assert d.get(11) == 1


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



import heapq
from copy import deepcopy
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy

from cityhash import CityHash32
from orso.profiler import distogram
from orso.schema import FlatColumn
from orso.types import OrsoTypes

MOST_FREQUENT_VALUE_SIZE: int = 32
KVM_SIZE: int = 32
INFINITY = float("inf")
SIXTY_FOUR_BITS: int = 8
SIXTY_FOUR_BYTES: int = 64
DISTOGRAM_BIN_COUNT: int = 50
MAX_INT64: int = 9223372036854775807


def string_to_int64(value: str) -> int:
    """Convert the first 8 characters of a string to an integer representation.

    Parameters:
        value: str
            The string value to be converted.

    Returns:
        An integer representation of the first 8 characters of the string.
    """
    byte_value = (value + "\x00\x00\x00\x00")[:SIXTY_FOUR_BITS].encode("utf-8")
    int_value = int.from_bytes(byte_value, "big")
    return int_value if int_value <= MAX_INT64 else MAX_INT64


def int64_to_string(value: int) -> str:
    # Convert the integer back to 8 bytes using big-endian byte order
    if value >= MAX_INT64:
        return None

    byte_value = value.to_bytes(SIXTY_FOUR_BITS, "big")

    # Decode the byte array back to a UTF-8 string
    # You might need to strip any padding characters that were added when encoding
    string_value = byte_value.decode("utf-8").rstrip("\x00")

    return string_value


def find_mfvs(data, top_n=MOST_FREQUENT_VALUE_SIZE):
    """
    Find the top N most frequent values (MFVs) in a NumPy array along with their counts.

    Parameters:
        data (np.ndarray): The input NumPy array containing numerical data.
        top_n (int): The number of top MFVs to return. Default is 32.

    Returns:
        top_values (np.ndarray): The top N most frequent values in the data.
        top_counts (np.ndarray): The counts of the top N most frequent values.
    """
    from collections import Counter

    counter = Counter(data)

    # Most common returns tuples of (value, count), so separate them
    top_items = counter.most_common(top_n)
    top_values, top_counts = zip(*top_items) if top_items else ([], [])

    return top_values, top_counts


def get_kvm_hashes(data, size: int):  # slowest function
    min_hashes = []

    data = list(set(data))

    # Build a list with the hash values of the first 'size' elements or all elements if fewer.
    min_hashes = [-CityHash32(str(element)) for element in data[:size]]

    # Transform the list into a heap in-place.
    heapq.heapify(min_hashes)

    for element in data[size:]:
        hash_value = CityHash32(str(element))

        # If the current hash is smaller than the largest in the heap
        if hash_value < -min_hashes[0]:
            heapq.heappushpop(min_hashes, -hash_value)

    # Convert back to positive values and sort before returning
    return sorted(-x for x in min_hashes)


def get_ordered_and_transitions(data) -> tuple:
    ordered = None
    transitions = 0
    last_value = data[0]
    for value in data[1:]:
        if value != last_value:
            transitions += 1
            if ordered is None:
                ordered = -1 if value < last_value else 1
            elif value > last_value and ordered == -1 or value < last_value and ordered == 1:
                ordered = 0
        last_value = value

    return (ordered, transitions)


@dataclass
class ColumnProfile:
    name: str
    type: OrsoTypes
    count: int = 0
    missing: int = 0
    maximum: Optional[int] = None
    minimum: Optional[int] = None
    order: int = None
    transitions: int = 0
    most_frequent_values: List[str] = field(default_factory=list)
    most_frequent_counts: List[int] = field(default_factory=list)
    histogram: List[Tuple] = field(default_factory=list)
    kmv_hashes: List[int] = field(default_factory=list)

    def deep_copy(self):
        """Create a deep copy of the Profile instance."""
        return deepcopy(self)

    def estimate_cardinality(self) -> int:
        """
        Estimates the cardinality (number of unique values) of the elements seen so far.
        """
        if not self.kmv_hashes:
            return 0
        # don't estimate if we know the number
        if len(self.kmv_hashes) < KVM_SIZE:
            return len(self.kmv_hashes)
        # Use the k-th smallest hash value (the last/largest in the sorted list) to estimate cardinality
        kth_min_value = self.kmv_hashes[-1]

        # Cardinality estimation formula
        return int((KVM_SIZE - 1) / (kth_min_value / 2**32))

    def estimate_values_at(self, point) -> int:
        if not hasattr(self, "distogram"):
            self.distogram = distogram.load(self.histogram, self.minimum, self.maximum)
        return distogram.bin_size(self.distogram, point)

    def estimate_values_below(self, point) -> int:
        if not hasattr(self, "distogram"):
            self.distogram = distogram.load(self.histogram, self.minimum, self.maximum)
        return distogram.count_at(self.distogram, point)

    def estimate_values_above(self, point) -> int:
        if not hasattr(self, "distogram"):
            self.distogram = distogram.load(self.histogram, self.minimum, self.maximum)
        return (self.count - self.missing) - distogram.count_at(self.distogram, point)

    def __add__(self, profile: "ColumnProfile") -> "ColumnProfile":
        new_profile = self.deep_copy()
        new_profile.count += profile.count
        new_profile.missing += profile.missing
        new_profile.transitions += profile.transitions + 1
        new_profile.order = 0 if new_profile.order == profile.order else new_profile.order
        new_profile.minimum = min([self.minimum or INFINITY, profile.minimum or INFINITY])
        if new_profile.minimum == INFINITY:
            new_profile.minimum = None
        new_profile.maximum = max([self.maximum or -INFINITY, profile.maximum or -INFINITY])
        if new_profile.maximum == -INFINITY:
            new_profile.maximum = None

        if self.most_frequent_values and profile.most_frequent_values:
            morsel1_map = dict(zip(self.most_frequent_values, self.most_frequent_counts))
            morsel2_map = dict(zip(profile.most_frequent_values, profile.most_frequent_counts))

            combined_map = {}
            for value in morsel1_map:
                if value in morsel2_map:  # Ensure the value is present in both morsels
                    combined_map[value] = morsel1_map[value] + morsel2_map[value]

            new_profile.most_frequent_values = (
                [] if len(combined_map) == 0 else list(combined_map.keys())
            )
            new_profile.most_frequent_counts = (
                [] if len(combined_map) == 0 else list(combined_map.values())
            )
        else:
            new_profile.most_frequent_values = []
            new_profile.most_frequent_counts = []

        if self.histogram and profile.histogram:
            my_dgram = distogram.load(self.histogram, self.minimum, self.maximum)
            profile_dgram = distogram.load(profile.histogram, profile.minimum, profile.maximum)
            if len(profile.histogram) > len(self.histogram):
                new_profile.histogram = distogram.merge(profile_dgram, my_dgram).bins
            else:
                new_profile.histogram = distogram.merge(my_dgram, profile_dgram).bins
        else:
            new_profile.histogram = []

        if self.kmv_hashes and profile.kmv_hashes:
            new_profile.kmv_hashes = sorted(set(self.kmv_hashes + profile.kmv_hashes))[:KVM_SIZE]

        return new_profile


class TableProfile:
    def __init__(self):
        self._columns: List[ColumnProfile] = []
        self._column_names: List[str] = []

    def __add__(self, right_profile: "TableProfile") -> "TableProfile":
        new_profile = TableProfile()

        for column_name in self._column_names:
            left_column = self.column(column_name)
            right_column = right_profile.column(column_name)
            if not right_column:
                right_column = ColumnProfile(
                    column_name, left_column.type, left_column.count, left_column.count
                )
            new_profile.add_column(left_column + right_column, column_name)

        return new_profile

    def add_column(self, profile: ColumnProfile, name: str):
        self._columns.append(profile)
        self._column_names.append(name)

    def __iter__(self):
        """An iterator over columns"""
        return iter(self.column)

    def column(self, i: Union[int, str]) -> Union[ColumnProfile, None]:
        """Get a column by its name or index"""
        if isinstance(i, str):
            for name, column in zip(self._column_names, self._columns):
                if name == i:
                    return column
            return None

        if isinstance(i, int):
            return self._columns[i]

    def to_dicts(self) -> List[dict]:
        return [asdict(v) for v in self._columns]

    def to_arrow(self) -> "pyarrow.Table":
        import pyarrow

        return pyarrow.Table.from_pylist(self.to_dicts())

    def to_dataframe(self) -> "DataFrame":
        import orso

        return orso.DataFrame(self.to_dicts())

    @classmethod
    def from_dataframe(cls, table) -> "TableProfile":
        from orso.schema import FlatColumn
        from orso.schema import RelationSchema

        profile = cls()

        profiler_classes = {
            OrsoTypes.VARCHAR: VarcharProfiler,
            OrsoTypes.INTEGER: NumericProfiler,
            OrsoTypes.DOUBLE: NumericProfiler,
            OrsoTypes.DECIMAL: NumericProfiler,
            OrsoTypes.ARRAY: ListStructProfiler,
            OrsoTypes.STRUCT: ListStructProfiler,
            OrsoTypes.BOOLEAN: BooleanProfiler,
            OrsoTypes.DATE: DateProfiler,
            OrsoTypes.TIMESTAMP: DateProfiler,
        }

        profiles = {}

        for morsel in table.to_batches(25000):
            if not isinstance(morsel.schema, RelationSchema):
                morsel._schema = RelationSchema(
                    name="morsel", columns=[FlatColumn(name=c) for c in morsel.schema]
                )

            for column in morsel.schema.columns:
                column_data = morsel.collect(column.name)
                if len(column_data) == 0:
                    continue

                profiler_class = profiler_classes.get(column.type, DefaultProfiler)
                profiler = profiler_class(column)
                profiler(column_data=column_data)
                if column.name in profiles:
                    profiles[column.name] += profiler.profile
                else:
                    profiles[column.name] = profiler.profile

        for name, summary in profiles.items():
            profile.add_column(summary, name)

        return profile


class BaseProfiler:
    def __init__(self, column: FlatColumn):
        self.column = column
        self.profile = ColumnProfile(name=column.name, type=column.type)

    def __call__(self, column_data: List[Any]):
        raise NotImplementedError("Must be implemented by subclass.")


class ListStructProfiler(BaseProfiler):
    def __call__(self, column_data: List[Any]):
        self.profile.count = len(column_data)
        self.profile.missing = sum(1 for val in column_data if val is None)


class DefaultProfiler(BaseProfiler):
    def __call__(self, column_data: List[Any]):
        self.profile.count = len(column_data)
        self.profile.missing = sum(1 for val in column_data if val != val)


class BooleanProfiler(BaseProfiler):
    def __call__(self, column_data: List[Any]):
        self.profile.count = len(column_data)

        column_data = [col for col in column_data if col is not None]
        self.profile.missing = self.profile.count - len(column_data)

        if len(column_data) > 0:
            self.profile.most_frequent_values = ["True", "False"]
            self.profile.most_frequent_counts = [column_data.count(True), column_data.count(False)]


class NumericProfiler(BaseProfiler):
    def __call__(self, column_data: List[Any]):
        self.profile.count = len(column_data)
        column_data = numpy.array(column_data, copy=False)  # Ensure column_data is a NumPy array
        if column_data.dtype.name == "object":
            column_data = column_data[~numpy.equal(column_data, -9223372036854775808)]
            column_data = [float(c) for c in column_data if c is not None]
        else:
            column_data = column_data[~numpy.isnan(column_data)]
        self.profile.missing = self.profile.count - len(column_data)
        # Compute min and max only if necessary
        if len(column_data) > 0:
            self.profile.minimum = int(numpy.min(column_data))
            self.profile.maximum = int(numpy.max(column_data))

            mf_values, mf_counts = find_mfvs(column_data, MOST_FREQUENT_VALUE_SIZE)
            self.profile.most_frequent_values = [f"{n:f}".rstrip("0").strip(".") for n in mf_values]
            self.profile.most_frequent_counts = mf_counts

            # Create a histogram of the data
            hist_counts, bin_edges = numpy.histogram(column_data, bins=DISTOGRAM_BIN_COUNT)
            self.profile.histogram = [
                (left_edge, count)
                for count, left_edge in zip(hist_counts, bin_edges[:-1])
                if count > 0
            ]

            # K-minimum value hashes, used for cardinality estimation
            self.profile.kmv_hashes = get_kvm_hashes(column_data, KVM_SIZE)
            self.profile.order, self.profile.transitions = get_ordered_and_transitions(column_data)


class VarcharProfiler(BaseProfiler):
    def __call__(self, column_data: List[Any]):
        self.profile.count = len(column_data)
        column_data = [col for col in column_data if col is not None]
        if len(column_data) > 0:
            # K-minimum value hashes, used for cardinality estimation
            self.profile.kmv_hashes = get_kvm_hashes(column_data, KVM_SIZE)
            column_data = [col[:SIXTY_FOUR_BYTES] for col in column_data]
            self.profile.missing = self.profile.count - len(column_data)
            self.profile.minimum = string_to_int64(min(column_data))
            self.profile.maximum = string_to_int64(max(column_data))

            mf_values, mf_counts = find_mfvs(column_data, MOST_FREQUENT_VALUE_SIZE)
            self.profile.most_frequent_values = mf_values
            self.profile.most_frequent_counts = mf_counts
            self.profile.order, self.profile.transitions = get_ordered_and_transitions(column_data)


class DateProfiler(BaseProfiler):
    def __call__(self, column_data: List[Any]):
        self.profile.count = len(column_data)
        if hasattr(column_data[0], "value"):
            column_data = numpy.array(
                [v.value for v in column_data if v is not None], dtype="int64"
            )
        else:
            column_data = numpy.array(column_data, dtype="datetime64[s]").astype("int64")
        column_data = column_data[~numpy.equal(column_data, -9223372036854775808)]
        self.profile.missing = self.profile.count - len(column_data)
        if len(column_data) > 0:
            numeric_profiler = NumericProfiler(self.column)
            numeric_profiler(column_data)
            numeric_profile = numeric_profiler.profile

            self.profile.minimum = numeric_profile.minimum
            self.profile.maximum = numeric_profile.maximum
            self.profile.most_frequent_values = numeric_profile.most_frequent_values
            self.profile.most_frequent_counts = numeric_profile.most_frequent_counts
            self.profile.histogram = numeric_profile.histogram
            self.profile.kmv_hashes = numeric_profile.kmv_hashes


def table_profiler(dataframe) -> List[Dict[str, Any]]:
    return TableProfile.from_dataframe(dataframe)



import os
import sys


sys.path.insert(1, os.path.join(sys.path[0], ".."))

import pytest
import time

# Import the LRU cache decorator
from orso.tools import lru_cache_with_expiry
from orso.tools import random_string

@lru_cache_with_expiry(max_size=3, valid_for_seconds=1)
def sample_function(x, y):
    return random_string()

def test_cache_basic_functionality():
    # Initial call, should not be cached
    result1 = sample_function(1, 2)

    # Call with same arguments, should be cached
    result2 = sample_function(1, 2)
    assert result2 == result1

def test_cache_expiry():
    # Initial call, should not be cached
    result1 = sample_function(2, 3)

    # Sleep for longer than the cache validity period
    time.sleep(1)

    # Call with same arguments, cache should have expired
    result2 = sample_function(2, 3)
    assert result2 != result1

def test_cache_lru_eviction():
    # Fill the cache with three different items
    result1 = sample_function(1, 2)  # Should be cached
    result2 = sample_function(2, 3)  # Should be cached
    result3 = sample_function(3, 4)  # Should be cached

    # At this point, the cache is full. The next item should cause the oldest item to be evicted
    result4 = sample_function(4, 5)  # Should be cached, (1, 2) should be evicted

    # Accessing the first item should result in a cache miss and recalculation
    result1_again = sample_function(1, 2)

    assert result1_again != result1  # New result as the old one should have been evicted

    # Ensure the recently accessed items are still cached
    result3_again = sample_function(3, 4)
    assert result3_again == result3


def test_cache_lru_access_order():
    # Fill the cache with three different items
    result1 = sample_function(5, 6)  # Should be cached
    result2 = sample_function(6, 7)  # Should be cached
    result3 = sample_function(7, 8)  # Should be cached

    # Access the first item to make it recently used
    result1_again = sample_function(5, 6)
    assert result1_again == result1

    # Add a new item to the cache, causing the LRU item to be evicted
    result4 = sample_function(8, 9)  # Should be cached, (6, 7) should be evicted as it is now the LRU

    # Accessing the evicted item should result in a cache miss and recalculation
    result2_again = sample_function(6, 7)
    assert result2_again != result2  # New result as the old one should have been evicted

if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



import os
import sys

import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.exceptions import DataValidationError
from orso.exceptions import ExcessColumnsInDataError
from orso.schema import RelationSchema
from orso.schema import FlatColumn
from orso.types import OrsoTypes
from tests import cities


def test_find_column():
    column = cities.schema.find_column("language")
    assert column.name == "language", column.name

    column = cities.schema.find_column("geolocation")
    assert column is None, column


def test_all_column_names():
    column_names = cities.schema.all_column_names()
    assert "name" in column_names
    assert "language" in column_names
    assert "geolocation" not in column_names


def test_schema_persistance():
    def strip_expectations(rel):
        columns = []
        for column in rel.columns:
            column.expectations = []
            columns.append(column)
        rel.columns = columns
        return rel

    as_dict = strip_expectations(cities.schema).to_dict()
    from_dict = strip_expectations(RelationSchema.from_dict(as_dict))

    assert strip_expectations(from_dict) == strip_expectations(cities.schema), strip_expectations(
        from_dict
    )
    assert as_dict == from_dict.to_dict()


def test_validate_with_valid_data():
    # Test with valid data
    data = {
        "name": "New York",
        "population": 8623000,
        "country": "United States",
        "founded": "1624",
        "area": 783.8,
        "language": "English",
    }
    assert cities.schema.validate(data) == True


def test_validate_with_missing_column():
    # Test with missing column
    data = {
        "name": "London",
        "population": 8908081,
        "country": "United Kingdom",
        "founded": "43 AD",
        "language": "English",
    }
    with pytest.raises(DataValidationError) as err:
        cities.schema.validate(data)
    assert "area" in str(err)


def test_validate_with_nullable_column():
    # Test with nullable column and value is None
    data = {
        "name": "Paris",
        "population": 2187526,
        "country": "France",
        "founded": None,
        "area": 105.4,
        "language": "French",
    }
    assert cities.schema.validate(data)


def test_validate_with_non_nullable_column():
    # Test with non-nullable column and value is None
    data = {
        "name": None,
        "population": 13929286,
        "country": "Japan",
        "founded": "1457",
        "area": 2187.66,
        "language": "Japanese",
    }
    with pytest.raises(DataValidationError) as err:
        cities.schema.validate(data)
    assert "name" in str(err)


def test_validate_with_wrong_type():
    # Test with column value of wrong type
    data = {
        "name": "Berlin",
        "population": 3769495,
        "country": "Germany",
        "founded": "1237",
        "area": "891.8",  # Expected type is double
        "language": "German",
    }
    with pytest.raises(DataValidationError) as err:
        cities.schema.validate(data)
    assert "area" in str(err)


def test_validate_with_multiple_errors():
    data = {
        "name": None,  # not nullable
        "population": 3769495,
        # country is missing
        "founded": "1237",
        "area": "891.8",  # Expected type is double
        "language": "German",
    }
    with pytest.raises(DataValidationError) as err:
        cities.schema.validate(data)
    assert "name" in str(err)
    assert "country" in str(err)
    assert "area" in str(err)


def test_validate_with_invalid_data_type():
    # Test with invalid data type (not a MutableMapping)
    data = [1, 2, 3]
    with pytest.raises(TypeError):
        cities.schema.validate(data)


def test_validate_with_additional_columns():
    # Test with column value of wrong type
    data = {
        "name": "Berlin",
        "population": 3769495,
        "country": "Germany",
        "founded": "1237",
        "area": "891.8",  # Expected type is double
        "language": "German",
        "continent": "Europe",
        "religion": "Christianity",
    }
    with pytest.raises(ExcessColumnsInDataError) as err:
        cities.schema.validate(data)

    # check both excess columns are in the error message
    assert "continent" in str(err)
    assert "religion" in str(err)


def test_schema_iterations():
    schema = cities.schema

    assert schema.num_columns == 6

    for i, column in enumerate(schema.columns):
        assert column == schema.column(i)
        assert column == schema.column(column.name)


def test_pop_column():
    # Clone the initial schema for the test to not alter the original
    schema = RelationSchema.from_dict(cities.schema.to_dict())

    # Check the initial state
    initial_column_count = len(schema.columns)

    # Test popping an existing column (e.g., "population")
    popped_column = schema.pop_column("population")
    assert popped_column is not None
    assert popped_column.name == "population"
    assert len(schema.columns) == initial_column_count - 1

    # Validate that the "population" column is no longer in the schema
    assert "population" not in [col.name for col in schema.columns]

    # Test popping a non-existent column
    popped_column = schema.pop_column("nonexistent")
    assert popped_column is None
    assert len(schema.columns) == initial_column_count - 1  # No change in length

    # Assert that the original columns, minus "population", are still there
    remaining_column_names = [col.name for col in schema.columns]
    assert "name" in remaining_column_names
    assert "country" in remaining_column_names
    assert "founded" in remaining_column_names
    assert "area" in remaining_column_names
    assert "language" in remaining_column_names


import pytest


def test_add_method_combines_columns():
    # Arrange
    col1 = FlatColumn(name="col1")
    col2 = FlatColumn(name="col2")
    col3 = FlatColumn(name="col3")

    schema1 = RelationSchema(name="Schema1", columns=[col1, col2])
    schema2 = RelationSchema(name="Schema2", columns=[col2, col3])

    # Act
    combined_schema = schema1 + schema2

    # Assert
    expected_columns = [col1, col2, col3]
    assert combined_schema.columns == expected_columns


def test_add_method_preserves_original_schemas():
    # Arrange
    col1 = FlatColumn(name="col1")
    col2 = FlatColumn(name="col2")
    col3 = FlatColumn(name="col3")
    schema1 = RelationSchema(name="Schema1", columns=[col1, col2])
    schema2 = RelationSchema(name="Schema2", columns=[col2, col3])

    # Act
    combined_schema = schema1 + schema2

    # Assert
    assert schema1.columns == [col1, col2]
    assert schema2.columns == [col2, col3]


def test_add_method_with_duplicate_columns():
    # Arrange
    col1 = FlatColumn(name="col1", type=0)
    col2 = FlatColumn(name="col2", type=0)
    col3 = FlatColumn(name="col3", type=0)
    schema1 = RelationSchema(name="Schema1", columns=[col1, col2])
    schema2 = RelationSchema(name="Schema2", columns=[col1, col2, col3])

    # Act
    combined_schema = schema1 + schema2

    # Assert
    expected_columns = [col1, col2, col3]
    assert combined_schema.columns == expected_columns


def test_minimum_definition():
    rs = RelationSchema.from_dict({"name": "relation", "columns": ["apples"]})
    assert len(rs.column_names) == 1
    assert rs.column("apples").type == OrsoTypes._MISSING_TYPE, rs.column("apples").type

    rs.validate({"apples": "none"})

    with pytest.raises(ExcessColumnsInDataError):
        rs.validate({"apples": "green", "oranges": "orange"})


def test_parsers():
    import datetime
    import decimal
    from orso.tools import parse_iso

    field = OrsoTypes.DATE
    parsed = field.parse("2023-01-01")
    assert isinstance(parsed, datetime.date), type(parsed)
    assert parsed == parse_iso("2023-01-01").date(), parsed

    field = OrsoTypes.DATE
    parsed = field.parse("2023-01-01T00:00:01")
    assert isinstance(parsed, datetime.date), type(parsed)
    assert parsed == parse_iso("2023-01-01").date(), parsed

    field = OrsoTypes.DATE
    with pytest.raises(Exception):
        parsed = field.parse("apples")

    field = OrsoTypes.DECIMAL
    parsed = field.parse("8.7")
    assert isinstance(parsed, decimal.Decimal), type(parsed)
    assert parsed == decimal.Decimal("8.7"), parsed

    field = OrsoTypes.DOUBLE
    parsed = field.parse("8.7")
    assert isinstance(parsed, float), type(parsed)
    assert parsed == 8.7, parsed

    field = OrsoTypes.DOUBLE
    parsed = field.parse("8")
    assert isinstance(parsed, float), type(parsed)
    assert parsed == 8.0, parsed

    field = OrsoTypes.INTEGER
    parsed = field.parse("8")
    assert isinstance(parsed, int), type(parsed)
    assert parsed == 8, parsed

    field = OrsoTypes.INTEGER
    parsed = field.parse("-8")
    assert isinstance(parsed, int), type(parsed)
    assert parsed == -8, parsed

    field = OrsoTypes.TIMESTAMP
    parsed = field.parse("2023-01-01T00:00:01")
    assert isinstance(parsed, datetime.datetime), type(parsed)
    assert parsed == parse_iso("2023-01-01 00:00:01"), parsed

    field = OrsoTypes.TIMESTAMP
    parsed = field.parse("1718530754")
    assert isinstance(parsed, datetime.datetime), type(parsed)
    assert parsed == parse_iso("2024-06-16 09:39:14"), parsed

    field = OrsoTypes.VARCHAR
    parsed = field.parse("1718530754")
    assert isinstance(parsed, str), type(parsed)
    assert parsed == "1718530754", parsed


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



# cython: infer_types=True
# cython: embedsignature=True
# cython: binding=False
# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: overflowcheck=False

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cpython.bytes cimport PyBytes_AsString, PyBytes_GET_SIZE
from cpython.object cimport PyObject_Str
from cython cimport int
from datetime import datetime
from ormsgpack import unpackb
from orso.exceptions import DataError
from typing import Dict, Any, Tuple
from libc.stdlib cimport malloc, free
import numpy as np
cimport cython
cimport numpy as cnp
from numpy cimport ndarray
from libc.stdint cimport int32_t

cnp.import_array()

HEADER_PREFIX = b"\x10\x00"
MAXIMUM_RECORD_SIZE = 8 * 1024 * 1024


cpdef from_bytes_cython(bytes data):
    cdef const char* data_ptr = PyBytes_AsString(data)
    cdef Py_ssize_t length = PyBytes_GET_SIZE(data)

    HEADER_SIZE = 14
    # Validate header and size, now using pointer arithmetic
    if length < HEADER_SIZE or (data_ptr[0] & 0xF0 != 0x10):
        raise DataError("Data malformed")
    
    # Deserialize record bytes
    cdef Py_ssize_t record_size = (
        (<unsigned char>data_ptr[2]) << 24 |
        (<unsigned char>data_ptr[3]) << 16 |
        (<unsigned char>data_ptr[4]) << 8  |
        (<unsigned char>data_ptr[5])
    )

    if record_size != length - HEADER_SIZE:
        raise DataError("Data malformed - incorrect length")

    # Deserialize and post-process
    cdef list raw_tuple = unpackb(data[HEADER_SIZE:])
    cdef list processed_list = []
    cdef object item


    for item in raw_tuple:
        if isinstance(item, list) and len(item) == 2 and item[0] == "__datetime__":
            processed_list.append(datetime.fromtimestamp(item[1]))
        else:
            processed_list.append(item)

    return tuple(processed_list)


cpdef tuple extract_dict_columns(dict data, tuple fields):
    cdef int i
    cdef str field
    cdef list field_data = [None] * len(fields)  # Preallocate list size

    for i, field in enumerate(fields):
        if field in data:
            field_data[i] = data[field]
    return tuple(field_data)  # Convert list to tuple


cpdef list collect_cython(list rows, cnp.ndarray[cnp.int32_t, ndim=1] columns, int limit=-1):
    """
    Collects columns from a list of tuples (rows).
    """
    cdef int32_t i, j, col_idx
    cdef int32_t num_rows = len(rows)
    cdef int32_t num_cols = columns.shape[0]
    cdef list row

    if limit >= 0 and limit < num_rows:
        num_rows = limit

    # Initialize result memory view with pre-allocated numpy arrays for each column
    cdef list result = [list([None] * num_rows) for _ in range(num_cols)]

    # Populate each column one at a time
    for j in range(num_cols):
        col_idx = columns[j]
        row = result[j]
        for i in range(num_rows):
            row[i] = rows[i][col_idx]
    
    # Convert each column back to a list and return the list of lists
    return result


cpdef int calculate_data_width(list column_values):
    cdef int width, max_width
    cdef object value
    
    max_width = 4  # Default width
    for value in column_values:
        if value is not None:
            width = PyBytes_GET_SIZE(PyObject_Str(value))
            if width > max_width:
                max_width = width

    return max_width


def process_table(table, row_factory, int max_chunksize):
    cdef list[] batches = table.to_batches(max_chunksize)
    cdef list rows = []
    cdef ndarray df_array

    for batch in batches:
        df_array = batch.to_pandas().replace({np.nan: None}).to_numpy()
        rows.extend(row_factory(i) for i in df_array)
    return rows



import datetime
import random

START_RANGE = datetime.datetime(1960, 1, 1)
END_RANGE = datetime.datetime(2100, 12, 31)


def generate_random_datetime(
    start: datetime.datetime = START_RANGE, end: datetime.datetime = END_RANGE
) -> datetime.datetime:
    """
    Generate a random datetime between two datetime objects.

    Parameters:
        start: datetime
            The start datetime.
        end: datetime
            The end datetime.

    Returns:
        datetime: A random datetime between the start and end.
    """
    delta = end - start
    random_seconds = random.randint(0, int(delta.total_seconds()))
    return start + datetime.timedelta(seconds=random_seconds)
